#include "gtk/gtk.h"
#include "mysql_table_editor_fe.h"
#include "linux_utilities/MGGladeXML.h"
#include "grtdb/db_object_helpers.h"
#include "treemodel_wrapper.h"
#include "gtk_helpers.h"
#include "auto_completable.h"

#include "mysql_table_editor_column_page.h"
#include "mysql_table_editor_column_flags.h"

#include "menu_manager.h"

#include <gtkmm/comboboxtext.h>
#include <gtkmm/textview.h>
#include <set>

AutoCompletable DbMySQLTableEditorColumnPage::_types_completion;
AutoCompletable DbMySQLTableEditorColumnPage::_names_completion;

//------------------------------------------------------------------------------
DbMySQLTableEditorColumnPage::DbMySQLTableEditorColumnPage(DbMySQLTableEditor *owner
                                                          ,MySQLTableEditorBE *be
                                                          ,MGGladeXML         *xml)
                             : _owner(owner)
                             , _be(be)
                             , _xml(xml)
                             , _edit_conn(0)
                             , _ce(0)
{
  _tv = _xml->get("table_columns", &_tv);
  _tv->set_enable_tree_lines(true);
  _tv->get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
  
  _model = ListModelWrapper::create(_be->get_columns(), _tv, "DbMySQLTableEditorColumnPage");

  _model->model().append_string_column(MySQLTableColumnsListBE::Name, "Column Name", EDITABLE, WITH_ICON);
  _model->model().append_combo_column(MySQLTableColumnsListBE::Type, "Datatype", _model->model().create_model(get_types_for_table(_be->table())), EDITABLE);
  _model->model().append_check_column(MySQLTableColumnsListBE::IsNotNull, "NN", EDITABLE);
  _model->model().append_check_column(MySQLTableColumnsListBE::IsAutoIncrement, "AI", EDITABLE);
  _model->model().append_string_column(MySQLTableColumnsListBE::Default, "Default", EDITABLE);

  _tv->set_model(_model);


  _types_completion.clear();
  std::vector<std::string> types(_be->get_columns()->get_datatype_names());
  
  for (std::vector<std::string>::const_iterator iter= types.begin(); iter != types.end(); ++iter)
  {
    if (*iter != "-")
      _types_completion.add_completion_text(*iter);
  }

  _names_completion.clear();
  _names_completion.add_completion_text("id");
  _names_completion.add_completion_text("name");
  _names_completion.add_completion_text(std::string(_be->table()->name().c_str()) + "_");
  
  // Add to completion list all column names in a schema
  std::set<std::string> column_names;
  db_mysql_SchemaRef schema = db_mysql_SchemaRef::cast_from(_be->get_table()->owner());
  if (schema.is_valid())
  {
    grt::ListRef<db_mysql_Table> tables = schema->tables();
    for (int i = tables.count() - 1; i >=0; --i)
    {
      grt::ListRef<db_mysql_Column> columns = tables[i]->columns();
      for (int j = columns.count() - 1; j >=0; --j)
      {
        column_names.insert(columns[j]->name().c_str());
      }
    }
  }
  
  std::set<std::string>::const_iterator it   = column_names.begin();
  std::set<std::string>::const_iterator last = column_names.end();
  for (; last != it; ++it)
    _names_completion.add_completion_text(*it);

  std::vector<Gtk::TreeViewColumn*> cols = _tv->get_columns();
  for (int j = cols.size() - 1; j >= 0; --j)
  {
    std::vector<Gtk::CellRenderer*> rends= cols[j]->get_cell_renderers();
    for (int i = rends.size() - 1; i >= 0; --i)
    {
      GtkCellRenderer* rend = rends[i]->gobj();
      rends[i]->set_data("idx", (gpointer)j);
      g_signal_connect(rend, "editing-started", GCallback(&DbMySQLTableEditorColumnPage::type_cell_editing_started), this);
    }
  }

  _tv->signal_event().connect(sigc::mem_fun(*this, &DbMySQLTableEditorColumnPage::process_event));
  _tv->signal_cursor_changed().connect(sigc::mem_fun(*this, &DbMySQLTableEditorColumnPage::cursor_changed));

  Gtk::TreeView *flags_view;
  _xml->get("column_flags", &flags_view);
  
  _column_flags = new ColumnFlags(flags_view);

  Gtk::TextView *column_comment;
  _xml->get("column_comment", &column_comment);
  _owner->add_text_change_timer(column_comment, sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::set_comment));

  _xml->get("column_collation_combo", &_collation_combo);
  _collation_combo->clear_items();
  fill_combo_from_string_list(_collation_combo, _be->get_charset_collation_list());
  _collation_combo->signal_changed().connect(sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::set_collation));

  _tv->signal_size_allocate().connect(sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::check_resize));
  _tv->signal_visibility_notify_event().connect(sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::do_on_visible));
}

//------------------------------------------------------------------------------
DbMySQLTableEditorColumnPage::~DbMySQLTableEditorColumnPage()
{
  delete _column_flags;
}

//--------------------------------------------------------------------------------
bec::NodeId DbMySQLTableEditorColumnPage::get_selected()
{
  Gtk::TreeModel::Path   path;
  Gtk::TreeView::Column *column(0);
  _tv->get_cursor(path, column);
  
  return _model->get_node_for_path(path);
}

//--------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::refresh()
{ 
  bec::ListModel* m = _model->get_be_model();
  _model->set_be_model(0);
  _tv->unset_model();

  _tv->set_model(_model);

  _model->set_be_model(m);
  _tv->unset_model();
  _model->refresh();
  _tv->set_model(_model);
}

//--------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::partial_refresh(const int what)
{
  switch (what)
  {
    case ::bec::TableEditorBE::RefreshColumnCollation:
    {
      update_collation();
      break;
    }
    default:;
  }
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::type_cell_editing_started(GtkCellRenderer* cr, GtkCellEditable* ce, gchar* path, gpointer udata)
{
  //Connect edit_done, so we can trigger refresh
  DbMySQLTableEditorColumnPage* columns_page = reinterpret_cast<DbMySQLTableEditorColumnPage*>(udata);

  const int idx = (int)((long long)gtk_object_get_data(GTK_OBJECT(cr), "idx"));
  
  bec::NodeId node(path);

  if ( GTK_IS_COMBO_BOX_ENTRY(ce) && idx == 1) // Attach types auto completion to the cell
  {
    GtkBin     *combo = GTK_BIN(ce);
    Gtk::Entry *entry = Glib::wrap((GtkEntry*)gtk_bin_get_child(combo));
    
    if ( entry )
      _types_completion.add_to_entry(entry);
  }
  else if ( GTK_IS_ENTRY(ce) && idx == 0) // Fill in name of the column
  {
    Gtk::Entry* entry = Glib::wrap(GTK_ENTRY(ce));

    std::string name;
    if (node.back() == columns_page->_be->get_columns()->count()-1)
      columns_page->_be->get_columns()->set_field(node, MySQLTableColumnsListBE::Name, 1);

    columns_page->_be->get_columns()->get_field(node, MySQLTableColumnsListBE::Name, name);
    entry->set_text(name);

    _names_completion.add_to_entry(entry);
  }


  // clean up edit_done signal/slot
  if ( columns_page->_ce && columns_page->_edit_conn )
  {
    g_signal_handler_disconnect (columns_page->_ce, columns_page->_edit_conn);
    columns_page->_ce = 0;
    columns_page->_edit_conn = 0;
  }

  if (GTK_IS_CELL_EDITABLE(ce))
  {
    columns_page->_ce = ce;
    columns_page->_edit_conn = g_signal_connect(ce, "editing-done", GCallback(&DbMySQLTableEditorColumnPage::cell_editing_done), udata);
  }
}

//--------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::cell_editing_done(GtkCellEditable* ce, gpointer udata)
{
  DbMySQLTableEditorColumnPage* columns_page = reinterpret_cast<DbMySQLTableEditorColumnPage*>(udata);

  if ( columns_page->_ce && columns_page->_edit_conn )
  {
    g_signal_handler_disconnect(columns_page->_ce, columns_page->_edit_conn);
    columns_page->_ce = 0;
    columns_page->_edit_conn = 0;
  }

  Gtk::TreeModel::Path   path;
  Gtk::TreeView::Column *column(0);
  columns_page->_tv->get_cursor(path, column);

  columns_page->refresh();

  columns_page->_tv->set_cursor(path);
}

//------------------------------------------------------------------------------
grt::StringListRef DbMySQLTableEditorColumnPage::get_types_for_table(const db_TableRef table)
{
  grt::StringListRef list(table->get_grt());
  std::vector<std::string> types(_be->get_columns()->get_datatype_names());
  
  for (std::vector<std::string>::const_iterator iter= types.begin(); iter != types.end(); ++iter)
  {
    if (*iter == "-")
      list.insert("----------");
    else
      list.insert(*iter);
  }

  return list;
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::check_resize(Gtk::Allocation& r)
{
  //   0       1        2  3         4
  // name    type      NN AI      Default
  const int step = r.get_width() / 10;
  _tv->get_column(0)->set_min_width(4 * step);
  _tv->get_column(1)->set_min_width(2 * step);
  _tv->get_column(4)->set_min_width(2 * step);
}

//------------------------------------------------------------------------------
bool DbMySQLTableEditorColumnPage::process_event(GdkEvent* event)
{
  if ( event->type == GDK_BUTTON_PRESS && event->button.button == 3 )
  {
    Gtk::TreeModel::Path   path;
    Gtk::TreeView::Column *column(0);
    int                    cell_x(-1);
    int                    cell_y(-1);
    
    if ( _tv->get_path_at_pos((int)event->button.x, (int)event->button.y, path, column, cell_x, cell_y) )
    {
      bec::NodeId node = _model->get_node_for_path(path);

      bec::MenuItemList menuitems;
      _be->get_columns()->get_node_popup_items(node, menuitems);
      MenuManager::run_popup(menuitems
                            ,event->button.x
                            ,event->button.y
                            ,event->button.time
                            ,sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::menu_action_on_node)
                            );
    }
  }
  else if ( event->type == GDK_2BUTTON_PRESS && event->button.button == 1 )
  {
    Gtk::TreeModel::Path   path;
    Gtk::TreeView::Column *column(0);
    int                    cell_x(-1);
    int                    cell_y(-1);
    
    if ( _tv->get_path_at_pos((int)event->button.x, (int)event->button.y, path, column, cell_x, cell_y) )
    {
      bec::NodeId node = _model->get_node_for_path(path);
      if ( node.is_valid() )
      {
        bool primary_key = _be->get_columns()->get_column_flag(node, "PRIMARY KEY");
        _be->get_columns()->set_column_flag(node, "PRIMARY KEY", !primary_key);
      }
    }
  }
  else if ( event->type == GDK_KEY_RELEASE )
  {
    type_column_event(event);
  }
  return false;
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::type_column_event(GdkEvent* event)
{
  if ( event->type == GDK_KEY_RELEASE )
  {
    const int key = event->key.keyval;
    if ( key == GDK_Tab/* || key == GDK_Return */)
    {
      // Advance to the next column
      Gtk::TreeModel::Path    path;
      Gtk::TreeView::Column  *column(0);
      _tv->get_cursor(path, column);
      
      if ( column )
      {
        //   0       1        2  3         4
        // name    type      NN AI      Default
        Glib::ListHandle<Gtk::TreeView::Column*> columns = _tv->get_columns();
  
        Glib::ListHandle<Gtk::TreeView::Column*>::const_iterator it   = columns.begin();
        Glib::ListHandle<Gtk::TreeView::Column*>::const_iterator last = columns.end();
        
        int i = 0;
        for ( ; last != it; ++it )
        {
          if ( (*it)->get_title() == column->get_title() )
            break;
          ++i;
        }

        refresh();
        
        ++it; ++i;
        if ( it != last && i <= 1 )
          _tv->set_cursor(path, **it, true);
        else // next row
        {
          path.next();
          _tv->set_cursor(path, **columns.begin(), true);
        }
      }
    }
  }
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::delete_column(const Gtk::TreeModel::Path& path)
{
  bec::NodeId node = _model->get_node_for_path(path);
  _be->remove_column(node);
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::reorder_column_up(const Gtk::TreeModel::Path& path)
{
  bec::NodeId node = _model->get_node_for_path(path);
  _be->get_columns()->reorder_up(node);
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::reorder_column_down(const Gtk::TreeModel::Path& path)
{
  bec::NodeId node = _model->get_node_for_path(path);
  _be->get_columns()->reorder_down(node);
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::menu_action_on_node(Gtk::MenuItem* item)
{
  const std::string item_name = item->get_name();
  
  std::list<Gtk::TreePath> selected = _tv->get_selection()->get_selected_rows();
  
  if ( item_name == "deleteSelectedColumnsToolStripMenuItem" )
    std::for_each(selected.begin(), selected.end(), sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::delete_column));
  else if ( item_name == "moveUpToolStripMenuItem" )
    std::for_each(selected.begin(), selected.end(), sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::reorder_column_up));
  else if ( item_name == "moveDownToolStripMenuItem" )
    std::for_each(selected.begin(), selected.end(), sigc::mem_fun(this, &DbMySQLTableEditorColumnPage::reorder_column_down));
  else if ( item_name == "refreshGridToolStripMenuItem" )
    refresh();
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::cursor_changed()
{
  Gtk::TreeModel::Path   path;
  Gtk::TreeView::Column *column(0);
  _tv->get_cursor(path, column);
  
  bec::NodeId node = _model->get_node_for_path(path);
  
  if ( node.is_valid() )
    update_column_details(node);
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::update_column_details(const ::bec::NodeId& node)
{
  Gtk::TreeView *flags_view;
  _xml->get("column_flags", &flags_view);
  
  _column_flags->fill(node, _be->get_columns(), flags_view);

  Gtk::TextView *column_comment;
  _xml->get("column_comment", &column_comment);
  
  std::string comment;
  _be->get_columns()->get_field(node, MySQLTableColumnsListBE::Comment, comment);
  
  column_comment->get_buffer()->set_text(comment);
  
  update_collation();
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::set_comment(const std::string& comment)
{
  const bec::NodeId node = get_selected();
  if (node.is_valid())
  {
    ::bec::TableColumnsListBE* cols = _be->get_columns();
    cols->set_field(node, (int)MySQLTableColumnsListBE::Comment, comment);
  }
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::set_collation()
{
  const bec::NodeId node = get_selected();
  if (node.is_valid())
  {
    bec::TableColumnsListBE *columns = _be->get_columns();
    columns->set_field(node, MySQLTableColumnsListBE::CharsetCollation, _collation_combo->get_active_text());
  }
}

//------------------------------------------------------------------------------
void DbMySQLTableEditorColumnPage::update_collation()
{
  const bec::NodeId node = get_selected();
  if (node.is_valid())
  {
    std::string has_charset;
    MySQLTableColumnsListBE* columns = _be->get_columns();
  
    Gtk::ComboBoxText *collation_combo;
    _xml->get("column_collation_combo", &collation_combo);
    
    columns->get_field(node, MySQLTableColumnsListBE::HasCharset, has_charset);
    if ( "1" == has_charset )
    {
      std::string column_cscoll;
      columns->get_field(node, MySQLTableColumnsListBE::CharsetCollation, column_cscoll);
      
      collation_combo->set_sensitive(true);
      collation_combo->set_active_text(column_cscoll);
    }
    else
    {
      collation_combo->set_sensitive(false);
    }
  }
}

//--------------------------------------------------------------------------------
bool DbMySQLTableEditorColumnPage::do_on_visible(GdkEventVisibility*)
{
  MySQLTableColumnsListBE *columns = _be->get_columns();
  if ( columns->count() == 1 )
  {
    ::bec::NodeId node(columns->get_node(0));
    _tv->set_cursor(node2path(node), *(_tv->get_column(0)), true);
  }
  return false;
}
