//
//  MySQLRoutineEditor.mm
//  MySQLWorkbench
//
//  Created by Alfredo Kojima on 22/Feb/09.
//  Copyright 2009 Sun Microsystems Inc. All rights reserved.
//

#import "MySQLRoutineEditor.h"
#import "MCPPUtilities.h"
#include "grtdb/db_object_helpers.h" // get_rdbms_for_db_object()

@implementation DbMysqlRoutineEditor

static void call_refresh(DbMysqlRoutineEditor *self)
{
  [self performSelectorOnMainThread:@selector(refresh) withObject:nil waitUntilDone:YES];
}


static void parser_log_callback(const std::vector<std::string> &strlist, DbMysqlRoutineEditor *self)
{
  std::string text;

  for (std::vector<std::string>::const_iterator line= strlist.begin(); line != strlist.end(); ++line)
    text.append(*line).append("\n");
  
  [self->errorText setToolTip: [NSString stringWithCPPString: text]];
  
  if (self->mBackEnd->has_syntax_error())
  {
    [self->errorImage setHidden: NO];
    [self->errorText setHidden: NO];
  }
  else
  {
    [self->errorImage setHidden: YES];
    [self->errorText setHidden: YES];
  }
  
  // name field could change...
  if (!self->mBackEnd->get_name().compare([[self->nameText stringValue] UTF8String]))
    [self->nameText setStringValue: [NSString stringWithCPPString:self->mBackEnd->get_name()]];
}


- (id)initWithModule:(grt::Module*)module GRTManager:(bec::GRTManager*)grtm arguments:(const grt::BaseListRef&)args
{
  self= [super initWithModule:module GRTManager:grtm arguments:args];
  if (self != nil)
  {
    // load GUI. Top level view in the nib is the NSTabView that will be docked to the main window
    if (![[NSBundle bundleForClass:[self class]] loadNibFile:@"MySQLRoutineEditor"
                                           externalNameTable:[NSDictionary dictionaryWithObject:self forKey:NSNibOwner] 
                                                    withZone:nil])
      NSLog(@"Could not load MySQLRoutineEditor.xib");
    
    // take the minimum size of the view from the initial size in the nib.
    // Therefore the nib should be designed as small as possible
    // note: the honouring of the min size is not yet implemented
    [self setMinimumSize: [tabView frame].size];
    
    // setup the editor backend with the schema object (args[0])
    mBackEnd= new MySQLRoutineEditorBE(grtm, db_mysql_RoutineRef::cast_from(args[0]), get_rdbms_for_db_object(args[0]));
    
    // register a callback that will make [self refresh] get called
    // whenever the backend thinks its needed to refresh the UI from the backend data (ie, the
    // edited object was changed from somewhere else in the application)
    mBackEnd->set_refresh_ui_slot(sigc::bind(sigc::ptr_fun(call_refresh), self));
    
    mBackEnd->set_sql_parser_log_cb(sigc::bind(sigc::ptr_fun(parser_log_callback), self));
    
    mNeedsInit= YES;
    
    // update the UI
    [self refresh];
  }
  return self;
}


- (void) dealloc
{
  [mTimer invalidate];
  [mTimer release];
  delete mBackEnd;
  [super dealloc];
}


- (BOOL)matchesIdentifierForClosingEditor:(NSString*)identifier
{
  return mBackEnd->should_close_on_delete_of([identifier UTF8String]);
}


/** Fetches object info from the backend and update the UI
 */
- (void)refresh
{
  if (mBackEnd)
  {
    std::string sql= mBackEnd->get_sql_definition_header();
    std::string code;
    
    [nameText setStringValue: [NSString stringWithCPPString:mBackEnd->get_name()]];
    
    code= mBackEnd->get_sql();
    
    if (code.empty())
    {
      int cursorPos;
      code= mBackEnd->get_sql_template("", cursorPos);
      sql.append(code);
      
      [errorImage setHidden: YES];
      [errorText setHidden: YES];
    }
    else
    {
      if (mBackEnd->has_syntax_error())
      {
        [errorImage setHidden: NO];
        [errorText setHidden: NO];
      }
      else
      {
        [errorImage setHidden: YES];
        [errorText setHidden: YES];
      }
      sql.append(code).append("//\n");
    }
  
    if (mNeedsInit)
      [codeText setString: [NSString stringWithCPPString:sql]];
    mNeedsInit= NO;
  }
}


- (id)identifier
{
  // an identifier for this editor (just take the object id)
  return [NSString stringWithCPPString:mBackEnd->get_object().id()];
}


- (NSString*)title
{
  // the title for the editor
  return [NSString stringWithFormat:@"%s - Routine", mBackEnd->get_name().c_str()];
}


- (NSView*)dockableView
{
  // the view to be docked to the main window
  return tabView;
}


- (void)updateBackend:(NSTimer*)theTimer
{
  [mTimer invalidate];
  [mTimer release];
  mTimer= 0;

  mBackEnd->set_sql([[codeText string] UTF8String]);
}



- (void)textDidChange:(NSNotification *)aNotification
{
  [mTimer invalidate];
  [mTimer release];
  
  mTimer= [NSTimer scheduledTimerWithTimeInterval:0.5
                                           target:self
                                         selector:@selector(updateBackend:) userInfo:nil
                                          repeats:NO];
  [mTimer retain];
}


- (void)textDidEndEditing:(NSNotification *)aNotification
{
  if ([aNotification object] == codeText)
  {    
    [self updateBackend:nil];
  }
}


@end
